/*
Model 819 test application.
Captures up to 3 streams per channel (primary and secondary H.264, JPEG).
Each stream is captured in a separate thread. A capture node
is created in each thread with all corresponding streams attached to it.
Capture can be enabled/disabled on a stream and channel basis.
H.264 streams are written to the files of limited length
overwriting old contents.
*/

#include <stdio.h>
#include <string.h>
#include <semaphore.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "s819.h"
#include "log.h"
#include "s819jpeg.h"			//test feature: JPEG recompression


//enable saving to files
#define __WRITE__
//log fole name
#define logfname				"log.txt"

#define MAX_CHANNELS            48      //max total input channels supported
#define MAX_STREAM_TYPES		3		//primary and secondary H.264, JPEG
#define MEGABYTE                (1024*1024)
#define JPEG_DIVIDER			10		//only save every 10-th captured JPEG frame
#define FDIVIDER				70		//divider for printing a character on screen
#define PRI						0		//primary H.264 stream
#define SEC						1		//secondary H.264 stream
#define JPG						2		//JPEG stream



// Globals
//////////////////////////////////////////////////////////////
int             gTotChans;          //number of channels in the system
int				gTermFlag = 0;		//termination flag
char			*streamName[MAX_STREAM_TYPES] = {"PRI", "SEC", "JPG"};	//stream type strings
int				maxFileLen[MAX_STREAM_TYPES] = {10*MEGABYTE, 3*MEGABYTE};	//max file sizes for PRI and SEC

//capture enable flags: 3 for each channel - primary, secondary, JPEG
char            chanEn[MAX_CHANNELS][3] = {
                                            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,

                                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

                                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};

//file handles
int             fh[MAX_CHANNELS][MAX_STREAM_TYPES];

//JPEG recompression quality
//At this point it's the same for all streams, set via command line argument
//Value of 0 means no recompression.
int				quality = 0;


///////////////////////////////////////////////////////////
//Helper functions
///////////////////////////////////////////////////////////

//Opens a file for output
//chan is 1-based
static int open_file (int chan, int stream)
{
    char                fname[64];
	char				strm[MAX_STREAM_TYPES] = {'p', 's'};
    int                 fh;

	if (stream == JPG) {    
		sprintf (fname, "j%02d.jpg", chan);
		fh = open (fname, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
	} else {
		sprintf (fname, "%c%02d.h264", strm[stream], chan);
		fh = open (fname, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
	}
    return fh;
}

//sigaction function: set global "terminate" flag
static void terminateAll (int sig, siginfo_t *si, void *unused)
{
	gTermFlag = 1;
}

//write to file function
static int write_file (int fh, char *buf, int length)
{
	int			ret;
	
	ret = write (fh, buf, length);
	return ret;
}

//find if particular stream capture enabled
static int isStreamEnabled (int stream)
{
	int			i;

	for (i = 0; i < gTotChans; i++) {
		if (chanEn[i][stream])
			return 1;
	}
	return 0;
}

///////////////////////////////////////////////////////////
//Capture thread function (H.264 primary and secondary
//Must be reentrant!
///////////////////////////////////////////////////////////
void *StreamCapThrFn (void *arg)
{
	HCNODE				hCnode;
	unsigned long	    stream = (unsigned long) arg;	//primary or sec
	int                 streams;        //stream enable mask
	int                 i, ret;
	ECODE               ecode;
	char				sym[MAX_STREAM_TYPES] = {'.', ':', 'x'};
	int                 nFrames[MAX_CHANNELS];    //total captured frames counter
	BUFFER              buffer;
	int					fileLen[MAX_CHANNELS];	//current file lengths
	int             	fh[MAX_CHANNELS];		//file handles
	int                 chan;           //channels # (1..gTotChans)
	JRHANDLE			jrh = NULL;		//handle to JPEG recompression object
	char				*jrOut;			//recompressed JPEG data;
	int					jrSize;			//recompressed JPEG size

	printf ("%s: %s thread started\n", __FUNCTION__, streamName[stream]);
	APPEND2LOG (logfname, "%s: Capturing %s streams\n", __func__, streamName[stream]);

	//Create capture node
    hCnode = S819_CreateCnode ();
    if (hCnode == NULL) {
        printf ("%s: Error creating CNODE for %s\n", __FUNCTION__, streamName[stream]);
        APPEND2LOG (logfname, "%s: Error creating CNODE for %s\n", __FUNCTION__, streamName[stream]);
        return NULL;
    } else {
        APPEND2LOG (logfname, "%s: Created CNODE for %s [%p]\n", __FUNCTION__, streamName[stream], hCnode);
	}		

    //Attach enabled streams to capture node.
    for (i = 0; i < gTotChans; i++) {
		nFrames[i] = FDIVIDER / gTotChans * i;	//create different offsets so that output is more uniform
        chan = i + 1;
        streams = 0;
        if (chanEn[i][stream]) {
            streams = SMASK(stream);	//stream must match S819_STREAM_TYPE values
        }

        if (streams) {				//if capture enabled
            ecode = S819_AttachStreams (hCnode, chan, streams);
            if (ecode != ECODE_OK) {
                printf ("%s: Error in S819_AttachStream %d\n", __FUNCTION__, ecode);
                APPEND2LOG (logfname, "%s: Error in S819_AttachStream %d\n", __FUNCTION__, ecode);
                goto cleanExit;
            }
			//open file for H.264 (JPG files are open each save)			
			if (stream != JPG) {
				fh[i] = open_file (chan, stream);
				if (fh[i] < 0) {
				    printf ("%s: File open error for stream %s\n", __FUNCTION__, streamName[stream]);
				    APPEND2LOG (logfname, "%s: File open error for stream %s\n", __FUNCTION__, streamName[stream]);
				    goto cleanExit;
				}
			}
		}
    }

    //Start all enabled streams
    ecode = S819_StartAll (hCnode);
    if (ecode != ECODE_OK) {
        printf ("%s: Error starting %s streams %d\n", __FUNCTION__, streamName[stream], ecode);
        APPEND2LOG (logfname, "%s: Error starting %s streams %d\n", __FUNCTION__, streamName[stream], ecode);
        return NULL;
    } else {
        printf ("%s: Stream started -- %s\n", __FUNCTION__, streamName[stream]);
        APPEND2LOG (logfname, "%s: Stream started -- %s\n", __FUNCTION__, streamName[stream]);
    }

	//Create JPEG recompressor if this is a JPEG stream and recompression is required
	if (stream == JPG && quality > 0) {
		jrh = S819_GetJRHandle (quality);
		if (jrh == NULL) {
		    printf ("%s: Failed to get JR handle\n", __FUNCTION__);
		    APPEND2LOG (logfname, "%s: Failed to get JR handle\n", __FUNCTION__);
		} else {
		    printf ("%s: Created a recompression object; quality %d\n", __FUNCTION__, quality);
		    APPEND2LOG (logfname, "%s: Created a recompression object; quality %d\n", __FUNCTION__, quality);
		}
	}

    //Capture loop
    while (!gTermFlag) {
        ecode = S819_WaitBuffer (hCnode, &buffer);
        if (ecode == ECODE_OK) {
			chan = buffer.chan - 1;			//0-based
            ++nFrames[chan];
            if (nFrames[chan] % FDIVIDER == 0) {
                printf ("%c", sym[stream]);
                fflush (stdout);
            }
#ifdef __WRITE__
			if (buffer.stream == STREAM_JPEG) {
				if (nFrames[chan] % JPEG_DIVIDER == 0) {		//only save every JPEG_DIVIDER-th JPEG frame 
					fh[chan] = open_file (chan + 1, stream);
					if (fh[chan] < 0) {
					    printf ("%s: File open error for stream %s\n", __FUNCTION__, streamName[stream]);
					    APPEND2LOG (logfname, "%s: File open error for stream %s\n", __FUNCTION__, streamName[stream]);
						goto cleanExit;
					}

					//Recompress JPEG if handle exists
					if (jrh) {
						S819_JRecompress (jrh, buffer.buf, &jrOut, &jrSize);
					    ret = write_file (fh[chan], jrOut, jrSize);
					} else {
					    ret = write_file (fh[chan], buffer.buf, buffer.length);
					}

				    if( ret < 0) {
				        fprintf( stderr,"Write to file err %d for stream %s\n", errno, streamName[stream]);
				        APPEND2LOG (logfname, "Write to file err %d for stream %s\n", errno, streamName[stream]);
				        goto cleanExit;
				    }
					close (fh[chan]);
				}					
			} else {		//H.264 stream
				if (fileLen[chan] >= maxFileLen[stream]) {	//if size limit reached
					fileLen[chan] = 0;						//reset file size
					lseek (fh[chan], 0, SEEK_SET);			//set file position to 0
				}
			
				//start file with H264_SPS slice so that it plays in VLC: skip non-SPS slices
				if (!((fileLen[chan] == 0) && (buffer.subtype.slice != H264_SPS))) {		
					ret = write_file (fh[chan], buffer.buf , buffer.length);
					if( ret < 0) {
						printf ("Write to file err %d for stream %s\n", errno, streamName[stream]);
						APPEND2LOG (logfname, "Write to file err %d for stream %s\n", errno, streamName[stream]);
						goto cleanExit;
					}
					fileLen[chan] += buffer.length;
				}
			}
#endif
		    ecode = S819_ReleaseBuffer (&buffer);
		    if (ecode != ECODE_OK) {
		        printf ("%s: S819_ReleaseBuffer error %d\n", __FUNCTION__, ecode);
		        APPEND2LOG (logfname, "%s: S819_ReleaseBuffer error %d\n", __FUNCTION__, ecode);
			}
        } else {
            printf ("%s: Acq error %d, stream %s\n", __FUNCTION__, ecode, streamName[stream]);
            APPEND2LOG (logfname, "%s: Acq error %d, stream %s\n", __FUNCTION__, ecode, streamName[stream]);
            continue;
        }
    }

cleanExit:
	printf ("%s: Stopping all %s streams\n", __FUNCTION__, streamName[stream]);
	APPEND2LOG (logfname, "%s: Stopping all %s streams\n", __FUNCTION__, streamName[stream]);
	ecode = S819_StopAll (hCnode);
    if (ecode != ECODE_OK) {
        printf ("%s: S819_StopAll error %d\n", __FUNCTION__, ecode);
        APPEND2LOG (logfname, "%s: S819_StopAll error %d\n", __FUNCTION__, ecode);
	}

	if (jrh)	
		S819_DeleteJRHandle (jrh);

	printf ("%s: Deleting capture node %s\n", __FUNCTION__, streamName[stream]);
	APPEND2LOG (logfname, "%s: Deleting capture node %s\n", __FUNCTION__, streamName[stream]);
	S819_DeleteCnode (hCnode);
	return NULL;
}



///////////////////////////////////////////////////////////
//main() function
///////////////////////////////////////////////////////////
int main (int argc ,char *argv[])
{
    struct sigaction 	actInt;
    struct sigaction 	actTerm;
    int             	i, j, ret;
    ECODE           	ecode;
    S819_CHAN       	s819chan[MAX_CHANNELS];
    pthread_t       	tid[MAX_STREAM_TYPES] = {0, 0, 0};
    void            	*status;
    unsigned long      	stream;

	//Get JPEG recompression quality from command line arg
	//100 - best, largest file, 0 - worst - smallest size

	if (argc > 1) {
		quality = atoi (argv[1]);
		if (quality < 0 || quality > 100) {
			printf ("Quality must be between 0 and 100.\n");
		}
	}

	//Initialize log
	ret = pthread_mutex_init (&logLock, NULL);
	if (ret) {
		printf ("Error initializing mutex\n");
		return ret;
	}
	RESETLOG (logfname);
	APPEND2LOG (logfname, "Log started\n");

    ecode = S819_Enumerate (&gTotChans);        //get # of chips
    if (ecode) {
        printf ("%s: Error enumerating %d\n", __FUNCTION__, ecode);
        return 0;
    }

    printf ("%s: Detected %d channels\n", __FUNCTION__, gTotChans);

	if (gTotChans == 0)
		goto errExit;

	//Set up Ctrl-C and terminate signal handlers
    actInt.sa_flags = SA_SIGINFO;
    actInt.sa_sigaction = terminateAll;
    sigaction (SIGINT, &actInt, NULL);


    actTerm.sa_flags = SA_SIGINFO;
    actTerm.sa_sigaction = terminateAll;
    sigaction (SIGTERM, &actTerm, NULL);

    //Set modes for all channels
    for (i = 1; i <= gTotChans; i++) {
        ecode = S819_SetMode (0, &s819chan[i - 1]);     //set defaults
        if (ecode != ECODE_OK) {
            printf ("%s: Error in S819_SetMode (defaults) %d\n", __FUNCTION__, ecode);
            goto errExit;
        }

        //make changes to mode here if required
		///////////////////////////////////////
        s819chan[i - 1].psMode.vMode.brCtrl = BITRATE_H264_VBR;
        s819chan[i - 1].psMode.vMode.bitrate = 2000000;
        s819chan[i - 1].psMode.vMode.interlace = MODE_INTERLACE;

        s819chan[i - 1].ssMode.vMode.brCtrl = BITRATE_H264_VBR;
        s819chan[i - 1].ssMode.vMode.bitrate = 500000;
        s819chan[i - 1].ssMode.vMode.interlace = MODE_PROGRESSIVE;

        s819chan[i - 1].jsMode.vMode.fps = 1;
        //////////////////////////////////////////

        ecode = S819_SetMode (i, &s819chan[i - 1]);     //set mode
        if (ecode != ECODE_OK) {
            printf ("%s: Error %d in S819_SetMode chan %d\n", __FUNCTION__, ecode, i);
            goto errExit;
        }
    }

	//Create threads for all captured streams
	stream = PRI;
	if (isStreamEnabled (stream)) {
		ret = pthread_create (&tid[PRI], NULL, &StreamCapThrFn, (void *)stream);
		if (ret < 0) {
		    printf ("%s: Error creating thread for %s\n", __FUNCTION__, streamName[PRI]);
		    goto errExit;
		}
	}

	stream = SEC;
	if (isStreamEnabled (stream)) {
		ret = pthread_create (&tid[SEC], NULL, &StreamCapThrFn, (void *)stream);
		if (ret < 0) {
		    printf ("%s: Error creating thread for %s\n", __FUNCTION__, streamName[SEC]);
		    goto errExit;
		}
	}

	stream = JPG;
	if (isStreamEnabled (stream)) {
		ret = pthread_create (&tid[JPG], NULL, &StreamCapThrFn, (void *)stream);
		if (ret < 0) {
		    printf ("%s: Error creating thread for %s\n", __FUNCTION__, streamName[JPG]);
		    goto errExit;
		}
	}

	//wait until all threads terminate
	for (i = 0; i < MAX_STREAM_TYPES; i++) {
		if (tid[i]) {
			ret = pthread_join (tid[i], (void **)&status);
		}
	}    	

errExit:
    printf ("%s: Calling S819_Close\n", __FUNCTION__);
    S819_Close();
    return 0;
}


